@@ -17,33 +17,27 @@ and if requested (using `gr=True`) the derivatives of the matching patterns
 in all primitives.
 """
 from math import sqrt
-from typing import Tuple, Union
+from typing import Literal, overload
 
 import numpy as np
 import scipy.linalg as spla
 
-from .utils import (
-    bs_error_abort,
-    der_nppow,
-    npexp,
-    npmaxabs,
-    nppow,
-    nprepeat_col,
-    nprepeat_row,
-    test_vector,
-)
+from .matching_utils import Matching
+from .utils import (bs_error_abort, der_nppow, npexp, npmaxabs, nppow,
+                    nprepeat_col, nprepeat_row, test_vector)
 
-TripleArrays = Tuple[np.ndarray, np.ndarray, np.ndarray]
-IPFPnoGradientResults = Tuple[TripleArrays, np.ndarray, np.ndarray, np.ndarray]
-IPFPGradientResults = Tuple[
-    TripleArrays, np.ndarray, np.ndarray, np.ndarray, TripleArrays
+ThreeArrays = tuple[np.ndarray, np.ndarray, np.ndarray]
+FourArrays = tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
+
+IPFPNoGradientResults = tuple[Matching, np.ndarray, np.ndarray]
+IPFPGradientResults = tuple[
+    Matching, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray
 ]
-IPFPResults = Union[IPFPnoGradientResults, IPFPGradientResults]
 
 
 def _ipfp_check_sizes(
     men_margins: np.ndarray, women_margins: np.ndarray, Phi: np.ndarray
-) -> Tuple[int, int]:
+) -> tuple[int, int]:
     """checks that the margins and surplus have the correct shapes and sizes"""
     X = test_vector(men_margins)
     Y = test_vector(women_margins)
@@ -52,15 +46,41 @@ def _ipfp_check_sizes(
     return X, Y
 
 
+@overload
 def ipfp_homoskedastic_nosingles_solver(
-    Phi: np.array,
-    men_margins: np.array,
-    women_margins: np.array,
+    Phi: np.ndarray,
+    men_margins: np.ndarray,
+    women_margins: np.ndarray,
+    tol: float,
+    gr: Literal[False],
+    verbose: bool,
+    maxiter: int,
+) -> ThreeArrays:
+    ...
+
+
+@overload
+def ipfp_homoskedastic_nosingles_solver(
+    Phi: np.ndarray,
+    men_margins: np.ndarray,
+    women_margins: np.ndarray,
+    tol: float,
+    gr: Literal[True],
+    verbose: bool,
+    maxiter: int,
+) -> FourArrays:
+    ...
+
+
+def ipfp_homoskedastic_nosingles_solver(
+    Phi: np.ndarray,
+    men_margins: np.ndarray,
+    women_margins: np.ndarray,
     tol: float = 1e-9,
     gr: bool = False,
     verbose: bool = False,
     maxiter: int = 1000,
-) -> IPFPResults:
+) -> ThreeArrays | FourArrays:
     """Solves for equilibrium in a Choo and Siow market without singles,
     given systematic surplus and margins
 
@@ -69,14 +89,14 @@ def ipfp_homoskedastic_nosingles_solver(
         men_margins: vector of men margins, shape (X)
         women_margins: vector of women margins, shape (Y)
         tol: tolerance on change in solution
-        gr: if `True`, also evaluate derivatives of $(\mu_{xy})$ wrt $\Phi$
+        gr: if `True`, also evaluate derivatives of $(\\mu_{xy})$ wrt $\\Phi$
         verbose: if `True`, prints information
         maxiter: maximum number of iterations
 
     Returns:
          muxy: the matching patterns, shape (X, Y)
          marg_err_x, marg_err_y: the errors on the margins
-         and the gradients of $(\mu_{xy})$ wrt $\Phi$ if `gr` is `True`
+         and the gradients of $(\\mu_{xy})$ wrt $\\Phi$ if `gr` is `True`
     """
     X, Y = _ipfp_check_sizes(men_margins, women_margins, Phi)
     n_couples = np.sum(men_margins)
@@ -140,7 +160,7 @@ def ipfp_homoskedastic_nosingles_solver(
         der_ephi2 /= 2.0 * ephi2  # 1/2 with safeguards
         ivar = 0
         for iman in range(X):
-            rhs[iman, ivar: (ivar + Y)] = -muxy[iman, :] * der_ephi2[iman, :]
+            rhs[iman, ivar : (ivar + Y)] = -muxy[iman, :] * der_ephi2[iman, :]
             ivar += Y
         ivar1 = X
         ivar2 = 0
@@ -157,7 +177,7 @@ def ipfp_homoskedastic_nosingles_solver(
         ivar = 0
         for iman in range(X):
             dt_man = dt[iman, :]
-            dmuxy[ivar: (ivar + Y), :] = np.outer((ephi2[iman, :] * tyi), dt_man)
+            dmuxy[ivar : (ivar + Y), :] = np.outer((ephi2[iman, :] * tyi), dt_man)
             ivar += Y
         for iwoman in range(Y):
             dT_woman = dT[iwoman, :]
@@ -171,14 +191,14 @@ def ipfp_homoskedastic_nosingles_solver(
 
 
 def ipfp_homoskedastic_solver(
-    Phi: np.array,
-    men_margins: np.array,
-    women_margins: np.array,
+    Phi: np.ndarray,
+    men_margins: np.ndarray,
+    women_margins: np.ndarray,
     tol: float = 1e-9,
     gr: bool = False,
     verbose: bool = False,
     maxiter: int = 1000,
-) -> IPFPResults:
+) -> IPFPNoGradientResults | IPFPGradientResults:
     """Solves for equilibrium in a Choo and Siow market with singles,
     given systematic surplus and margins
 
@@ -275,7 +295,11 @@ def ipfp_homoskedastic_solver(
         print(f"\tMargin error on x: {npmaxabs(marg_err_x)}")
         print(f"\tMargin error on y: {npmaxabs(marg_err_y)}")
     if not gr:
-        return (muxy, mux0, mu0y), marg_err_x, marg_err_y
+        return (
+            Matching(muxy, men_margins, women_margins),
+            marg_err_x,
+            marg_err_y,
+        )
     else:  # we compute the derivatives
         sxi = ephi2 @ tyi
         syi = ephi2T @ txi
@@ -298,7 +322,7 @@ def ipfp_homoskedastic_solver(
         der_ephi2 /= 2.0 * ephi2  # 1/2 with safeguards
         ivar = n_sum_categories
         for iman in range(X):
-            rhs[iman, ivar: (ivar + Y)] = -muxy[iman, :] * der_ephi2[iman, :]
+            rhs[iman, ivar : (ivar + Y)] = -muxy[iman, :] * der_ephi2[iman, :]
             ivar += Y
         ivar1 = X
         ivar2 = n_sum_categories
@@ -317,7 +341,7 @@ def ipfp_homoskedastic_solver(
         ivar = 0
         for iman in range(X):
             dt_man = dt[iman, :]
-            dmuxy[ivar: (ivar + Y), :] = np.outer((ephi2[iman, :] * tyi), dt_man)
+            dmuxy[ivar : (ivar + Y), :] = np.outer((ephi2[iman, :] * tyi), dt_man)
             ivar += Y
         for iwoman in range(Y):
             dT_woman = dT[iwoman, :]
@@ -327,19 +351,54 @@ def ipfp_homoskedastic_solver(
         # add the term that comes from differentiating ephi2
         muxy_vec2 = (muxy * der_ephi2).reshape(n_prod_categories)
         dmuxy[:, n_sum_categories:] += np.diag(muxy_vec2)
-        return (muxy, mux0, mu0y), marg_err_x, marg_err_y, (dmuxy, dmux0, dmu0y)
+        return (
+            Matching(muxy, men_margins, women_margins),
+            marg_err_x,
+            marg_err_y,
+            dmuxy,
+            dmux0,
+            dmu0y,
+        )
+
+
+@overload
+def ipfp_gender_heteroskedastic_solver(
+    Phi: np.ndarray,
+    men_margins: np.ndarray,
+    women_margins: np.ndarray,
+    tau: float,
+    tol: float,
+    gr: Literal[False],
+    verbose: bool,
+    maxiter: int,
+) -> IPFPNoGradientResults:
+    ...
+
+
+@overload
+def ipfp_gender_heteroskedastic_solver(
+    Phi: np.ndarray,
+    men_margins: np.ndarray,
+    women_margins: np.ndarray,
+    tau: float,
+    tol: float,
+    gr: Literal[True],
+    verbose: bool,
+    maxiter: int,
+) -> IPFPGradientResults:
+    ...
 
 
 def ipfp_gender_heteroskedastic_solver(
-    Phi: np.array,
-    men_margins: np.array,
-    women_margins: np.array,
+    Phi: np.ndarray,
+    men_margins: np.ndarray,
+    women_margins: np.ndarray,
     tau: float,
     tol: float = 1e-9,
     gr: bool = False,
     verbose: bool = False,
     maxiter: int = 1000,
-) -> IPFPResults:
+) -> IPFPNoGradientResults | IPFPGradientResults:
     """Solves for equilibrium in a in a gender-heteroskedastic Choo and Siow market
     given systematic surplus and margins and a scale parameter `tau`
 
@@ -372,7 +431,14 @@ def ipfp_gender_heteroskedastic_solver(
     tau_y = np.full(Y, tau)
 
     if gr:
-        mus_hxy, marg_err_x, marg_err_y, dmus_hxy = ipfp_heteroskedastic_solver(
+        (
+            mus_hxy,
+            marg_err_x,
+            marg_err_y,
+            dmus_xy,
+            dmus_x0,
+            dmus_0y,
+        ) = ipfp_heteroskedastic_solver(
             Phi,
             men_margins,
             women_margins,
@@ -383,8 +449,7 @@ def ipfp_gender_heteroskedastic_solver(
             maxiter=maxiter,
             verbose=verbose,
         )
-        muxy, mux0, mu0y = mus_hxy
-        dmus_xy, dmus_x0, dmus_0y = dmus_hxy
+        muxy, _, _, _, _ = mus_hxy.unpack()
         n_sum_categories = X + Y
         n_prod_categories = X * Y
         n_cols = n_sum_categories + n_prod_categories
@@ -398,7 +463,14 @@ def ipfp_gender_heteroskedastic_solver(
         dmu0y = np.zeros((Y, n_cols + 1))
         dmu0y[:, :n_cols] = dmus_0y[:, :n_cols]
         dmu0y[:, -1] = np.sum(dmus_0y[:, itau_y:], 1)
-        return (muxy, mux0, mu0y), marg_err_x, marg_err_y, (dmuxy, dmux0, dmu0y)
+        return (
+            Matching(muxy, men_margins, women_margins),
+            marg_err_x,
+            marg_err_y,
+            dmuxy,
+            dmux0,
+            dmu0y,
+        )
 
     else:
         return ipfp_heteroskedastic_solver(
@@ -414,17 +486,47 @@ def ipfp_gender_heteroskedastic_solver(
         )
 
 
+@overload
 def ipfp_heteroskedastic_solver(
-    Phi: np.array,
-    men_margins: np.array,
-    women_margins: np.array,
-    sigma_x: np.array,
-    tau_y: np.array,
+    Phi: np.ndarray,
+    men_margins: np.ndarray,
+    women_margins: np.ndarray,
+    sigma_x: np.ndarray,
+    tau_y: np.ndarray,
+    tol: float,
+    gr: Literal[False],
+    verbose: bool,
+    maxiter: int,
+) -> IPFPNoGradientResults:
+    ...
+
+
+@overload
+def ipfp_heteroskedastic_solver(
+    Phi: np.ndarray,
+    men_margins: np.ndarray,
+    women_margins: np.ndarray,
+    sigma_x: np.ndarray,
+    tau_y: np.ndarray,
+    tol: float,
+    gr: Literal[True],
+    verbose: bool,
+    maxiter: int,
+) -> IPFPGradientResults:
+    ...
+
+
+def ipfp_heteroskedastic_solver(
+    Phi: np.ndarray,
+    men_margins: np.ndarray,
+    women_margins: np.ndarray,
+    sigma_x: np.ndarray,
+    tau_y: np.ndarray,
     tol: float = 1e-9,
     gr: bool = False,
     verbose: bool = False,
     maxiter: int = 1000,
-) -> IPFPResults:
+) -> IPFPNoGradientResults | IPFPGradientResults:
     """Solves for equilibrium in a in a fully heteroskedastic Choo and Siow market
     given systematic surplus and margins
     and standard errors `sigma_x` and `tau_y`
@@ -537,7 +639,11 @@ def ipfp_heteroskedastic_solver(
         print(f"\tMargin error on x: {npmaxabs(marg_err_x)}")
         print(f"\tMargin error on y: {npmaxabs(marg_err_y)}")
     if not gr:
-        return (muxy, mux0, mu0y), marg_err_x, marg_err_y
+        return (
+            Matching(muxy, men_margins, women_margins),
+            marg_err_x,
+            marg_err_y,
+        )
     else:  # we compute the derivatives
         n_sum_categories = X + Y
         n_prod_categories = X * Y
@@ -584,7 +690,7 @@ def ipfp_heteroskedastic_solver(
         #  to compute derivatives of (mux0, mu0y) wrt Phi
         ivar = n_sum_categories
         for iman in range(X):
-            rhs[iman, ivar: (ivar + Y)] = -big_a[iman, :]
+            rhs[iman, ivar : (ivar + Y)] = -big_a[iman, :]
             ivar += Y
         ivar1 = X
         ivar2 = n_sum_categories
@@ -614,7 +720,7 @@ def ipfp_heteroskedastic_solver(
         der1 = ephi2 * der_axy1 * bxy
         ivar = 0
         for iman in range(X):
-            dmuxy[ivar: (ivar + Y), :] = np.outer(der1[iman, :], dmux0[iman, :])
+            dmuxy[ivar : (ivar + Y), :] = np.outer(der1[iman, :], dmux0[iman, :])
             ivar += Y
         der2 = ephi2 * der_bxy1 * axy
         for iwoman in range(Y):
@@ -635,7 +741,7 @@ def ipfp_heteroskedastic_solver(
         ivar = 0
         ix = iend_phi
         for iman in range(X):
-            dmuxy[ivar: (ivar + Y), ix] -= big_c[iman, :]
+            dmuxy[ivar : (ivar + Y), ix] -= big_c[iman, :]
             ivar += Y
             ix += 1
         # on the derivative wrt tau_y
@@ -644,4 +750,11 @@ def ipfp_heteroskedastic_solver(
             dmuxy[iwoman:n_prod_categories:Y, iy] -= big_d[:, iwoman]
             iy += 1
 
-        return (muxy, mux0, mu0y), marg_err_x, marg_err_y, (dmuxy, dmux0, dmu0y)
+        return (
+            Matching(muxy, men_margins, women_margins),
+            marg_err_x,
+            marg_err_y,
+            dmuxy,
+            dmux0,
+            dmu0y,
+        )
